local L = AceLibrary("AceLocale-2.2"):new("SanityItemCache")

SanityItemCache = AceLibrary("AceAddon-2.0"):new("AceConsole-2.0", "AceEvent-2.0", "AceDB-2.0", "AceDebug-2.0")
SanityItemCache:RegisterDB("SanityItemCacheDB", "SanityItemCacheDBPC")

local QualityStrings = {L["Trash"], L["Common"], L["Uncommon"], L["Rare"], L["Epic"], L["Legendary"]}

local default_player_setup = {
	Bags = {},
	BagInfo = {},
	ItemCategories = {},
	CategoryCategories = {},
	ItemTags = {},
	Money = 0
}

local function translate_index(self, k)
	local keys = SanityItemCache.translationKeys
	-- We really don't want to blow our stack on valid numeric indexes
	if tonumber(k) then
		setmetatable(self,nil)
		local v = self[k]
		setmetatable(self, SanityItemCache.translateMeta)
		return v		
	end
	
	if keys[k] then
		return self[keys[k]]
	elseif k ~= "Link" and self[3] and (
		k == "Name" or k == "ID" or k == "Quality" or k == "iLevel" or k == "MinLevel" or k == "Type" or
		k == "SubType" or k == "Stack" or k == "EquipLoc" or k == "Texture") then
		local n = GetItemInfo(self[3])
		if not n then
			self.invalid = true
		end
		local name, idstr, itemQuality, itemLevel, itemMinLevel, itemType, subType, stack, equipLoc, texture  = GetItemInfo(self[3]);
		if k == "Name" then
			self[k] = name or "Unknown"
			return self[k]
		elseif k == "ID" then
			self[k] = idstr or -1
			return self[k]
		elseif k == "Quality" then
			self[k] = itemQuality or 1
			return self[k]
		elseif k == "iLevel" then
			self[k] = itemLevel or 0
			return self[k]
		elseif k == "MinLevel" then
			self[k] = itemMinLevel or 0
			return self[k]
		elseif k == "Type" then
			self[k] = itemType or "Unknown"
			return self[k]
		elseif k == "SubType" then
			self[k] = subType or "Unknown"
			return self[k]
		elseif k == "Stack" then
			self[k] = stack or 1
			return self[k]
		elseif k == "EquipLoc" then
			self[k] = SANITY_CATEGORYNAMES[equipLoc] or equipLoc or L["INVALID"]
			return self[k]
		elseif k == "Texture" then
			self[k] = texture or "Interface\\Icons\\INV_Misc_QuestionMark"
			return self[k]
		end
	end
	return nil
end

SanityItemCache:RegisterDefaults("realm", { chars = {} })

----[[ Start Category Handling ]]----

function SanityItemCache:SetItemCategory(itemName, categoryName)
	SanityItemCache.db.profile.ItemCategories[itemName] = categoryName
	SanityItemCache:UnprepItem(item)
end

function SanityItemCache:DeleteCategory()
end

function SanityItemCache:SetCategoryCategory(itemName, categoryName)
end

function SanityItemCache:GetExistingCategories()
	local existingCategoryNames = {}
	local existingCategories = {}
	local i = 0
	for k,v in pairs(SanityItemCache.db.profile.ItemCategories) do
		if v then
			existingCategoryNames[v] = 1
		end
	end
	for k,v in pairs(SanityItemCache.db.profile.CategoryCategories) do
		if v then
			existingCategoryNames[v] = 1
		end
	end
	local cats = {}
	for k,v in pairs(existingCategoryNames) do
		tinsert(cats, k)
	end
	return cats
end
----[[ Start Tagging ]]----

function SanityItemCache:ItemHasTag(item, tag)
	if self.db.profile.ItemTags[item] then
		return self.db.profile.ItemTags[item][tag] ~= nil
	end
	return false
end

function SanityItemCache:GetCategoryFor(item)
	local name = self.db.profile.ItemCategories[item]
	if name then
		local catName = self.db.profile.CategoryCategories[name]
		if catName then return catName end
	end
	return name
end

function SanityItemCache:AddItemTag(item, tag)
	if not self.db.profile.ItemTags[item] then
		self.db.profile.ItemTags[item] = {}
	end
	self.db.profile.ItemTags[item][tag] = 1
	self:UnprepItem(item)
	self:GetTagList(true)
end

function SanityItemCache:RemoveItemTag(item, tag, noRefresh)
	if not self.db.profile.ItemTags[item] then return end
	local tTags = {}
	for k,v in pairs(self.db.profile.ItemTags[item]) do
		if k ~= tag then
			tTags[k] = 1
		end
	end
	self.db.profile.ItemTags[item] = tTags
	self:UnprepItem(item)
	self:GetTagList(true)
end

function SanityItemCache:GetItemsForTag(tag)
	if not self.db.profile.ItemTags then 
		return
	end
	local result = {}
	local resultCount = 0
	for k,v in pairs(self.db.profile.ItemTags) do
		for k2, v2 in pairs(v) do
			if k2 == tag then
				tinsert(result, k)
			end
		end
	end
	return result
end

function SanityItemCache:GetTagsFor(item)
	if not self.db.profile.ItemTags[item] then return "" end
	local s = ""
	local tags = {}
	for k,v in pairs(self.db.profile.ItemTags[item]) do
		tinsert(tags, k)
	end
	table.sort(tags)
	for k,v in pairs(tags) do
		if s == "" then
			s = v
		else
			s = s .. ", " .. v
		end
	end
	return s
end

function SanityItemCache:KillTag(tag)
	if not self.db.profile.ItemTags then return end
	for k,v in pairs(self.db.profile.ItemTags) do
		self:UntagItem(k, tag)
	end
	self:GetTagList(true)
end

function SanityItemCache:GetTagList(force)
	if self.tagList and not self.tagListDirty and not force then return self.tagList end
	self.lastTagList = self.lastTagList or {}
	for i = 1, #self.lastTagList do
		tremove(self.lastTagList)
	end
	local tags = self.lastTagList
	for k,v in pairs(self.db.profile.ItemTags) do
		for k1,v1 in pairs(v) do
			tags[k1] = 1
		end
	end
	self.tagList = self.tagList or {}
	local tList = self.tagList
	for i = 1, #self.tagList do
		tremove(self.tagList)
	end
	for k,v in pairs(tags) do
		tinsert(tList, tostring(k))
	end
	table.sort(tList)
	self.tagListDirty = false
	return tList
end

----[[ End Tagging ]]----

function SanityItemCache:OnInitialize()
	self.char = UnitName("player")
	self:RegisterEvent("BAG_UPDATE")
	self:RegisterEvent("PLAYER_MONEY")
	self:RegisterEvent("BANKFRAME_OPENED")
	self:RegisterEvent("BANKFRAME_CLOSED")
	self:RegisterEvent("UNIT_INVENTORY_CHANGED")
	self:RegisterEvent("ADDON_LOADED")
	self:RegisterEvent("PLAYER_LOGOUT")
	self.translationKeys = {["Count"] = 1, ["Bag"] = 2, ["Link"] = 3, ["Soulbound"] = 4, ["Slot"] = 5}
	self.translateMeta = {__index = translate_index}
	self.BankIsVisible = false
	
	SanityItemCache_ItemTooltip = CreateFrame("GameTooltip", "SanityItemCache_ItemTooltip", WorldFrame, "GameTooltipTemplate")
	SanityItemCache_ItemTooltip:SetOwner(WorldFrame, "ANCHOR_NONE")

	self.db.profile.ItemCategories = self.db.profile.ItemCategories or {}
	self.db.profile.CategoryCategories = self.db.profile.CategoryCategories or {}
	self.db.profile.ItemTags = self.db.profile.ItemTags or {}

	if not self.db.realm.chars[self.char] then
		self.db.realm.chars[self.char] = {}
		for k,v in pairs(default_player_setup) do
			self.db.realm.chars[self.char][k] = {}
		end
	end
	local upgraded = false
	for k,v in pairs(self.db.realm.chars) do
		if v.BagItems then
			upgraded = true
			v.BagItems = nil
			v.MailItems = nil
			v.EquipmentItems = nil
			v.BankItems = nil
			v.Bags = {}
			v.BagInfo = {}
		end
		if v.Bags and type(v.Bags) == "table" and #v.Bags > 0 then
			for k2,v2 in pairs(v.Bags) do
				if type(v2) ~= "table" then
					v.BagItems = nil
					v.MailItems = nil
					v.EquipmentItems = nil
					v.BankItems = nil
					v.Bags = {}
					v.BagInfo = {}
					upgraded = true
					break
				end
			end
		end
		v.BagInfo = v.BagInfo or {}
	end
	if upgraded then
		SanityItemCache:Print("Please note that Sanity has undergone a database reorganization and your old item data is no longer available. Please check item locations for each character you wish to have item data for.")
	end
	self.INVENTORY_BAG_ID = -99
	self.MAILBOX_BAG_ID = -98
	SanityItemCache:PLAYER_MONEY()
	self.CacheInited = true
end

function SanityItemCache:GetBagNameFor(bagID)
	if bagID == self.INVENTORY_BAG_ID then return L["Wearing"] end
	if bagID == self.MAILBOX_BAG_ID then return L["Mailbox"] end
	
	if bagID == -1 then return L["Bank"] end
	if bagID == 5 then return L["Bank"] end
	if bagID == 6 then return L["Bank"] end
	if bagID == 7 then return L["Bank"] end
	if bagID == 8 then return L["Bank"] end
	if bagID == 9 then return L["Bank"] end
	if bagID == 10 then return L["Bank"] end
	if bagID == 11 then return L["Bank"] end
	
	if bagID == 0 then return L["Bags"] end
	if bagID == 1 then return L["Bags"] end
	if bagID == 2 then return L["Bags"] end
	if bagID == 3 then return L["Bags"] end
	if bagID == 4 then return L["Bags"] end
	
	if bagID == KEYRING_CONTAINER then return L["Keyring"] end
	
	return "INVALID"
end

function SanityItemCache:NukeCharacter(char)
	local tt = {}
	for tchar,dummy in pairs(self.db.realm.chars) do
		if tchar ~= char then
			tt[tchar] = dummy
		end
	end
	self.db.realm.chars = tt	
end

function SanityItemCache:ADDON_LOADED()
	local upgraded = false
	for k,v in pairs(self.db.realm.chars) do
		if v.BagItems then
			upgraded = true
			v.BagItems = nil
			v.MailItems = nil
			v.EquipmentItems = nil
			v.BankItems = nil
			v.Bags = {}
			v.BagInfo = {}
		end
		if v.Bags and type(v.Bags) == "table" and #v.Bags > 0 then
			for k2,v2 in pairs(v.Bags) do
				if type(v2) ~= "table" then
					v.BagItems = nil
					v.MailItems = nil
					v.EquipmentItems = nil
					v.BankItems = nil
					v.Bags = {}
					v.BagInfo = {}
					upgraded = true
					break
				end
			end
		end
		v.BagInfo = v.BagInfo or {}
	end
	if upgraded then
		SanityItemCache:Print("Please note that Sanity has undergone a database reorganization and your old item data is no longer available. Please check item locations for each character you wish to have item data for.")
	end
	
	for char,dummy in pairs(self.db.realm.chars) do
		for k,v in pairs(dummy.Bags) do
			for k2,v2 in pairs(v) do
				setmetatable(v2,{__index = translate_index})
			end
		end
	end
	SanityItemCache:ScanAll()
end

function SanityItemCache:PLAYER_LOGOUT()
	for charName, char in pairs(self.db.realm.chars) do
		for k,v in pairs(char.Bags) do
			local t = {}
			for k2,v2 in pairs(v) do
				tinsert(t, {
					[1] = v2[1],
					[2] = v2[2],
					[3] = v2[3],
					[4] = v2[4],
					[5] = v2[5]
				})
			end
			char.Bags[k] = t
		end
	end
end

function SanityItemCache:ItemIsUsable(item)
	if UnitName("player") == item["Owner"] then
		if item["Bag"] == self.MAILBOX_BAG_ID then return false end		-- We can't use items in the mail
		if item["Bag"] == self.INVENTORY_BAG_ID then return true end	-- We can always use items we're wearing
		if item["Bag"] >= 0 and item["Bag"] <= 4 then return true end	-- We can always use items in our bags
		return self.BankIsVisible
	end
	return false
end

function SanityItemCache:BANKFRAME_OPENED()
	self.BankIsVisible = true
	SanityItemCache:ScanBag(-1)
	for i = 5, 11 do
		SanityItemCache:ScanBag(i)
	end
end

function SanityItemCache:BANKFRAME_CLOSED()
	self.BankIsVisible = false
end

function SanityItemCache:UNIT_INVENTORY_CHANGED()
	self.db.realm.chars[self.char].Bags[self.INVENTORY_BAG_ID], self.db.realm.chars[self.char].BagInfo[self.INVENTORY_BAG_ID] = SanityItemCache:ScanEquipment(), {0,0}
end

function SanityItemCache:ScanAll()
	SanityItemCache:ScanBags()
	SanityItemCache:UNIT_INVENTORY_CHANGED()
end

function SanityItemCache:BAG_UPDATE()
	if not tonumber(arg1) then return end
	SanityItemCache:ScanBag(arg1)
	if self.BankIsVisible then
		SanityItemCache:ScanBag(-1)
		for i = 5, 11 do
			SanityItemCache:ScanBag(i)
		end
	end
end

function SanityItemCache:UnprepItem(itemName)
	for k, bag in pairs(self.db.realm.chars[self.char].Bags) do
		for k, item in pairs(bag) do
			if item["Name"] == itemName then
				item["Prepped"] = false
			end
		end
	end
end

function SanityItemCache:PLAYER_MONEY()
	self.db.realm.chars[self.char].Money = GetMoney()
end

function SanityItemCache:ScanBag(bagID)
	self.countDirty = true
	local scan = true
	-- NB - Make sure we're doing bag filtering display-side
	--[[
	if bagID >= 0 and bagID <= 4 then
		scan = Sanity.db.profile["ScanBag" .. bagID]
	elseif bagID == KEYRING_CONTAINER then
		scan = Sanity.db.profile.ScanKeyRing
	end
	]]--
	if scan then
		SanityItemCache:Debug("Scanning bag " .. bagID)
		local a, b = SanityItemCache:Scan(bagID)
		if a then
			self.db.realm.chars[self.char].Bags[bagID], self.db.realm.chars[self.char].BagInfo[bagID] = a, b
		end
	end
end

function SanityItemCache:GetBagData(char)
	local bags = {0, 1, 2, 3, 4}
	if not self.CacheInited then return 0,0 end
	return self:CountSlots(bags, char)
end

function SanityItemCache:GetBankData(char)
	local bags = {-1, 5, 6, 7, 8, 9, 10, 11}
	if not self.CacheInited then return 0,0 end
	return self:CountSlots(bags)
end
	
function SanityItemCache:CountSlots(bags, char)
	if not char then char = self.char end
	local c, cf = 0, 0
	for k, bag in ipairs(bags) do
		if self.db.realm.chars[char].BagInfo[bag] then
			c = c + self.db.realm.chars[char].BagInfo[bag].bag_slots
			cf = cf + self.db.realm.chars[char].BagInfo[bag].filled_slots
		end
	end
	return cf, c
end

function SanityItemCache:ScanBags()
	for i = 0, 4 do
		-- NB - make sure we're doing bag filtering display-side
		-- if Sanity.db.profile["ScanBag" .. i] then
			self:ScanBag(i)
		-- end
	end
	-- if Sanity.db.profile.ScanKeyRing then
		self:ScanBag(KEYRING_CONTAINER)
	-- end
end

local function tokenize(str, split, t)
	for i = 1, #t do
		tremove(t)
	end
	t[1], t[2], t[3], t[4], t[5], t[6], t[7], t[8], t[9], t[10], t[11], t[12], t[13], t[14], t[15], t[16], t[17], t[18], t[19], t[20] = strsplit(split, str)
end

function SanityItemCache:MatchExpr(item, s, ns)
	if not item then return false end
	if not item.Name then return false end

	self.tokenTable = self.tokenTable or {}

	s = s:gsub(": ", ":"):gsub(" :", ":")
	self:PrepareItem(item)
	if not item.Name then
		SanityItemCache:PrintLiteral(item)
		return
	end
	
	local expr, remainder = strsplit(",", s)
	while expr ~= nil do
		local m = 0
		local t = 0
		local invert = false
		expr = expr:trim()
		local token, remainder2 = strsplit(" ", expr)
		while token ~= nil do
			local match = false
			t = t + 1
			token = token:gsub("\"","")
			local a, b = strsplit(":", token:lower(), 2)
			if not b then
				if a == "*" then
					match = not match
				else
					if a:sub(1, 1) == "!" then
						match = not match
						a = a:sub(2)
					end
					if item.Name:lower():find(a) then match = not match
					elseif item.Tags:lower():find(a) then match = not match
					elseif item.TypeSort:lower():find(a) then match = not match
					elseif QualityStrings[item.Quality + 1]:lower():find(a) then match = not match
					end
				end
			else
				if b:sub(1, 1) == "!" then
					match = not match
					b = b:sub(2)
				end
				if a == "name" and item.Name:lower():find(b) then match = not match
				elseif (a == "tags" or a == "tag") and item.Tags:lower():find(b) then match = not match
				elseif (a == "bag") and item.Bag == tonumber(b) then match = not match
				elseif (a == "cat" or a == "category" or a == "type") and item.TypeSort:lower():find(b) then match = not match
				elseif (a == "quality" or a == "color" or a == "rarity") and QualityStrings[item.Quality+1]:lower() == b then match = not match
				elseif (a == "tooltip" or a == "desc") and item.Tooltip:lower():find(b) then match = not match
				elseif (a == "soulbound" and (b == "true" or b == "t" or b == "1") and item.Soulbound == 1) or (a == "soulbound" and not (b == "true" or b == "t" or b == "1") and item.Soulbound ~= 1) then match = not match
				end
			end
			if match then
				if ns then
					if item.Name:lower():find(ns) then
						m = m + 1
					end
				else
					m = m + 1
				end
			end
			if not remainder2 then break end
			token, remainder2 = strsplit(" ", remainder2)
		end
		if m == t then return true end
		if not remainder then break end
		expr, remainder = strsplit(",", remainder)
	end
	return false
end

function SanityItemCache:Scan(bag)
	local items = {}
	local bagData = {bag_slots = 0, filled_slots = 0}
	if GetContainerNumSlots(bag) == 0 then return nil end
	bagData.bag_slots = GetContainerNumSlots(bag)
	for slot = 1, GetContainerNumSlots(bag) do
		local link = GetContainerItemLink(bag, slot)
		if link then
			if bag ~= KEYRING_CONTAINER then
				bagData.filled_slots = (bagData.filled_slots or 0) + 1
			end
			local n, _, itemQuality, itemLevel, itemMinLevel, itemType, subType, stack, equipLoc, texture  = GetItemInfo(link);
			if itemQuality then
				SanityItemCache_ItemTooltip:ClearLines()
				if bag == -1 then
					SanityItemCache_ItemTooltip:SetInventoryItem("player", BankButtonIDToInvSlotID(slot, nil));
				else
					SanityItemCache_ItemTooltip:SetBagItem(bag, slot)
				end

				local soulbound = 0
				if bag ~= KEYRING_CONTAINER then
					for i=1,SanityItemCache_ItemTooltip:NumLines() do
						local obj = getglobal("SanityItemCache_ItemTooltipTextLeft" .. i)
						if obj:GetText() == L["Soulbound"] or obj:GetText() == L["Quest Item"] then
							soulbound = 1
						end
					end
				else
					soulbound = 1
				end
				
				local texture, count, locked = GetContainerItemInfo(bag, slot);
				local _,_,itemID = string.find(link, "H(item:%-?%d+:%-?%d+:%-?%d+:%-?%d+:%-?%d+:%-?%d+:%-?%d+:%-?%d+)")
				local item = {count, bag, itemID, soulbound, slot}
				setmetatable(item,{__index = translate_index})
				item = self:PrepareItem(item)
				tinsert(items, item)
			end
		end
	end
	return items, bagData
end

function SanityItemCache:ScanEquipment()
	local items = {}
	for i = 1,19 do
		SanityItemCache_ItemTooltip:SetInventoryItem("player", i)
		local soulbound = 0
		for i=1,SanityItemCache_ItemTooltip:NumLines() do
			if SanityItemCache_ItemTooltipTextLeft2:GetText() == L["Soulbound"] or SanityItemCache_ItemTooltipTextLeft2:GetText() == L["Quest Item"] then
				soulbound = 1
			end
		end
		local link = GetInventoryItemLink("player", i)
		if link then
			local name = GetItemInfo(link)
			local _,_,itemID = string.find(link, "H(item:%-?%d+:%-?%d+:%-?%d+:%-?%d+:%-?%d+:%-?%d+:%-?%d+:%-?%d+)")
			local item = {GetInventoryItemCount("player", i), SanityItemCache.INVENTORY_BAG_ID, itemID, soulbound, i}
			setmetatable(item,{__index = translate_index})
			item = self:PrepareItem(item)
			tinsert(items, item)
		end
	end
	return items
end

function SanityItemCache:ScanMailbox()
	local inbox_items = GetInboxNumItems();
	local items = {}
	
	for i = 1, GetInboxNumItems() do
		--[[ Sanity_ItemTooltip:SetMailboxItem("player", i)
		local soulbound = 0
		for i=1,Sanity_ItemTooltip:NumLines() do
			if Sanity_ItemTooltipTextLeft2:GetText() == L["Soulbound"] or Sanity_ItemTooltipTextLeft2:GetText() == L["Quest Item"] then
				soulbound = 1
			end
		end
		]]--
		local soulbound = 0
		local link = GetInboxItemLink(i)
		if link then
			local itemName, itemLink, itemRarity, itemLevel, itemMinLevel, itemType, itemSubType, itemStackCount, itemEquipLoc, invTexture = GetItemInfo(link)
			local _,_,itemID = string.find(link, "H(item:%-?%d+:%-?%d+:%-?%d+:%-?%d+:%-?%d+:%-?%d+:%-?%d+:%-?%d+)")
			SanityItemCache:Debug(itemID)
			local name, itemTexture, count, quality, canUse = GetInboxItem(i);
			local item = {count, SanityItemCache.MAILBOX_BAG_ID, itemID, soulbound, i}
			setmetatable(item,{__index = translate_index})
			item = self:PrepareItem(item)
			tinsert(items, item)
		end
	end
	self.db.realm.chars[self.char].Bags[self.MAILBOX_BAG_ID], self.db.realm.chars[self.char].BagInfo[self.MAILBOX_BAG_ID] = items, {0,0}
	-- return items
end

function SanityItemCache:GetTooltipString(link)
	if not link then return nil end
-- 	if not Sanity.db.profile.SearchTooltips then return nil end
	local name = GetItemInfo(link)
	if not name then name = "Unknown" end
	self.TooltipDataCache = self.TooltipDataCache or {}
	if self.TooltipDataCache[name] then
		return self.TooltipDataCache[name]
	end
	SanityItemCache_ItemTooltip:ClearLines()
	SanityItemCache_ItemTooltip:SetOwner(WorldFrame, "ANCHOR_NONE")
	SanityItemCache_ItemTooltip:SetHyperlink(link)
	local s = nil
	for i=1,SanityItemCache_ItemTooltip:NumLines() do
		local lText = getglobal("SanityItemCache_ItemTooltipTextLeft" .. i)
		local rText = getglobal("SanityItemCache_ItemTooltipTextRIght" .. i)
		if lText and lText:GetText() then
			if s then s = s .. " " .. lText:GetText() 
			else s = lText:GetText() end
		end
		if rText and rText:GetText() then
			if s then s = s .. " " .. rText:GetText() 
			else s = rText:GetText() end
		end
	end
	self.TooltipDataCache[name] = s
	return s
end

function SanityItemCache:PrepareItem(item)
	if item["Prepped"] then return end
	item.SetCategory = function(catName)
		self.SetItemCategory(self["Name"], cat)
	end
	item.AddTag = function(tag)
		self.AddItemTag(self["Name"], tag)
	end

	item.RemoveTag = function(tag)
		self.RemoveItemTag(self["Name"], tag)
	end

	local skip = false
	if not item["Prepped"] then
		item["Location"] = self:GetBagNameFor(item["Bag"])
		item["Prepped"] = true
		if not skip then
			item["Stacks" ] = 1
			if item["Link"] then
				if item["Dangerous"] then
					item["Dangerous"] = false
				end	
				if not item["Type"] then
					item["Type"] = L["Unknown"]
					item["SubType"] = L["Unknown"]
					item["EquipLoc"] = L["Unknown"]
					item["Dangerous"] = true
				end
			end	
			item["Tags"] = SanityItemCache:GetTagsFor(item["Name"])
			if not item.invalid then
				item["Tooltip"] = SanityItemCache:GetTooltipString(item["Link"]) or ""
			else
				item["Tooltip"] = ""
			end
			item["TypeSort"] = "zzz"
			local catName = SanityItemCache.db.profile.ItemCategories[item["Name"]]
			if catName then
				item["TypeSort"] = catName
			else
				if item["Type"] == item["SubType"] then
					item["TypeSort"] = item["SubType"]
					if item["EquipLoc"] and #item["EquipLoc"] > 0 then
						item["TypeSort"] = item["TypeSort"] .. " - " .. item["EquipLoc"]
					end
				elseif item["EquipLoc"] and #item["EquipLoc"] > 0 then
					item["TypeSort"] = item["Type"] .. " - " .. item["EquipLoc"]
				else
					item["TypeSort"] = item["Type"]
				end
			end
			local catCatName = SanityItemCache.db.profile.CategoryCategories[item["TypeSort"]]
			if catCatName then
				item["TypeSort"] = catCatName
			end
			if item["TypeSort"] == nil then
				item["TypeSort"] = "No Type"
			end
		end
	end
	return item
end